package main

import (
	"context"
	"embed"
	"flag"
	"net/http"
	"os"
	"os/signal"
	"time"

	kitlog "github.com/go-kit/kit/log"
	"github.com/go-kit/kit/log/level"
	_ "modernc.org/ql/driver"
	_ "modernc.org/sqlite"

	"git.c3pb.de/gbe/invinoveritas/auth"
	"git.c3pb.de/gbe/invinoveritas/log"
	"git.c3pb.de/gbe/invinoveritas/session"
	"git.c3pb.de/gbe/invinoveritas/storage"
	"git.c3pb.de/gbe/invinoveritas/storage/query"
)

//go:embed templates/*.tpl
var templateFS embed.FS

//go:embed static/*
var staticFS embed.FS

// Build info. Will be set by Gitlab pipeline.
var (
	commitHash string
	buildTime  string
)

func httpError(w http.ResponseWriter, r *http.Request, msg string, err error, status int) {
	if err != nil {
		msg += ": " + err.Error()
	}

	level.Error(log.Get(r)).
		Log("status", status,
			"error", err,
			"msg", msg)

	http.Error(w, msg, status)
}

type sessionProvider interface {
	ListSessions(context.Context) ([]session.Info, error)
	DeleteSession(ctx context.Context, token string) error
}

type userProvider interface {
	ListUsers(context.Context) ([]string, error)
	CreateUser(ctx context.Context, name string) (string, error)
	DeleteUser(ctx context.Context, name string) error

	UpdatePassword(ctx context.Context, name string, oldPW, newPW string, createNew bool) error
}

type Handler struct {
	sp sessionProvider
	up userProvider

	Q *query.Queries
}

func addCacheHeaders(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		w.Header().Add("Cache-Control", "public, max-age=86400, immutable")
		next.ServeHTTP(w, r)
	})
}

// wrapMiddleware wraps common middleware around hdlr:
// - log.Request to make a logger available in the request context
// - addCacheHeader for caching
// - auth.Require
func wrapMiddleware(hdlr http.Handler, sessions session.Provider, logger kitlog.Logger) http.Handler {
	authFailed := log.Request(sessions.Handler(templateFS), logger)
	return log.Request(auth.Require(hdlr, authFailed, sessions), logger)
}

func main() {
	dbPath := flag.String("db", "vino.sqlite", "Path to database file")
	listenAddr := flag.String("listen", "127.0.0.1:7878", "Listening address")
	debug := flag.Bool("debug", false, "Enable debug logging")

	flag.Parse()

	logger := kitlog.NewLogfmtLogger(kitlog.NewSyncWriter(os.Stdout))
	logger = kitlog.With(logger, "ts", kitlog.DefaultTimestampUTC)

	filter := level.AllowInfo()
	if *debug {
		filter = level.AllowDebug()
	}

	logger = level.NewFilter(logger, filter)

	level.Info(logger).
		Log("commit", commitHash, "build", buildTime)

	ctx, done := context.WithCancel(context.Background())
	defer done()

	db, err := storage.Open(ctx, *dbPath, logger)
	if err != nil {
		level.Error(logger).Log("error", err, "msg", "can't open DB")
		os.Exit(1)
	}
	defer db.Close()

	logged404 := log.Request(http.HandlerFunc(http.NotFound), kitlog.With(logger, "code", 404))

	http.Handle("/favicon.ico", logged404)

	http.Handle("/robots.txt", log.Request(addCacheHeaders(robots()), logger))
	http.Handle("/static/", log.Request(addCacheHeaders(http.FileServer(http.FS(staticFS))), logger))

	sessions, err := session.NewProvider(ctx, query.New(db))
	if err != nil {
		level.Error(logger).Log("error", err, "msg", "can't initialize session provider")
		os.Exit(1)
	}

	handler := Handler{
		sp: sessions,
		up: sessions,

		Q: query.New(db),
	}

	http.Handle("/details/img/", addCacheHeaders(wrapMiddleware(handler.img(), sessions, logger)))
	http.Handle("/details/", wrapMiddleware(handler.details(), sessions, logger))

	http.Handle("/user/sessions", wrapMiddleware(handler.user(pageSessions), sessions, logger))
	http.Handle("/user/all-users", wrapMiddleware(handler.user(pageAllUsers), sessions, logger))
	http.Handle("/user/you", wrapMiddleware(handler.user(pageYou), sessions, logger))

	index := wrapMiddleware(handler.index(), sessions, logger)

	http.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
		if r.URL.Path != "/" && r.URL.Path != "/add" {
			logged404.ServeHTTP(w, r)
			return
		}

		index.ServeHTTP(w, r)
	})

	srv := http.Server{
		Addr: *listenAddr,
	}

	level.Info(logger).
		Log("addr", "http://"+*listenAddr, "msg", "starting http server")

	errs := make(chan error, 1)

	go func() {
		errs <- http.ListenAndServe(*listenAddr, nil)
	}()

	// Wait for OS signal or error, shut down server if signal received
	// TODO: Use oklog/run here instead.
	sigChan := make(chan os.Signal, 1)
	signal.Notify(sigChan, os.Interrupt)

	select {
	case sig := <-sigChan:
		level.Info(logger).
			Log("signal", sig, "msg", "received signal, terminating")

		shutdownCtx, done := context.WithTimeout(context.Background(), 15*time.Second)
		defer done()

		err = srv.Shutdown(shutdownCtx)
		if err != nil {
			level.Error(logger).
				Log("error", err, "msg", "clean shut down failed")
		}
	case err = <-errs:
		level.Error(logger).
			Log("error", err, "msg", "HTTP server failed to start")
	}
}
