package session

import (
	"context"
	crand "crypto/rand"
	"crypto/sha256"
	"database/sql"
	"errors"
	"fmt"
	"html/template"
	"io/fs"
	"math/rand"
	"net/http"
	"sync"
	"time"

	kitlog "github.com/go-kit/kit/log"
	"github.com/go-kit/kit/log/level"

	"git.c3pb.de/gbe/invinoveritas/auth"
	"git.c3pb.de/gbe/invinoveritas/log"
	"git.c3pb.de/gbe/invinoveritas/storage/query"
)

func hashPassword(ctx context.Context, q *query.Queries, user, pass string) (string, error) {
	// Look up password salt from DB and hash password with it
	salt, err := q.GetSalt(ctx)
	if err != nil {
		return "", fmt.Errorf("loading password salt: %w", err)
	}

	h := sha256.New()

	_, err = h.Write([]byte(salt))
	if err != nil {
		return "", err
	}

	fmt.Fprint(h, ":", user, ":", pass)

	return fmt.Sprintf("%02x", h.Sum(nil)), nil
}

type Info struct {
	Name    string    `db:"name"`
	Created time.Time `db:"created"`
	Remote  string    `db:"remote"`
}

type Provider struct {
	q *query.Queries
}

func NewProvider(ctx context.Context, q *query.Queries) (Provider, error) {
	// Check if there's a password seed in the DB already. If not, add one.
	err := q.RunTx(ctx, func(q *query.Queries) error {
		_, err := q.GetSalt(ctx)
		if err != nil && !errors.Is(err, sql.ErrNoRows) {
			return err
		}
		if err == nil {
			return nil
		}

		var rawSalt [40]byte

		n, err := crand.Read(rawSalt[:])
		if err != nil {
			return err
		}
		if n < len(rawSalt) {
			return errors.New("short read")
		}

		err = q.InsertSalt(ctx, fmt.Sprintf("%02x", rawSalt))
		if err != nil {
			return err
		}

		return nil
	})
	if err != nil {
		return Provider{}, err
	}

	return Provider{
		q: q,
	}, nil
}

// Valid looks up the provided session token in a's database and returns the appropriate user if the token is
// valid. If it is not, an error is returned.
func (a Provider) Valid(ctx context.Context, token string) (*auth.User, error) {
	var (
		user auth.User
		err  error
	)

	err = a.q.RunTx(ctx, func(q *query.Queries) error {
		user.Name, err = q.IsValidSession(ctx, token)
		if errors.Is(err, sql.ErrNoRows) {
			// Let's see if there are any users at all. If not, we let 'em in.
			count, err := q.CountUsers(ctx)
			if err != nil {
				return err
			}

			// No users in DB, allow everyone
			if count == 0 {
				return nil
			}

			return auth.ErrAuthFailed
		}

		return nil
	})
	if err != nil {
		level.Error(log.GetContext(ctx)).
			Log("error", err, "token", token, "msg", "got error during session lookup")
		return nil, err
	}

	return &user, nil
}

// Handler returns an HTTP handler that can be used to ask a user to authenticate themselves. It loads templates
// from the given FS.
func (a Provider) Handler(templateFS fs.FS) http.Handler {
	var (
		once sync.Once
		tpl  *template.Template
	)

	type responseData struct {
		Error string
	}

	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		once.Do(func() {
			tpl = template.Must(template.ParseFS(templateFS, "templates/base.tpl", "templates/auth.tpl"))
		})

		if r.Method == "GET" {
			err := tpl.ExecuteTemplate(w, "auth.tpl", nil)
			if err != nil {
				level.Error(log.Get(r)).
					Log("error", err, "msg", "can't execute auth template")
			}

			return
		}

		// Get user and password from POST data
		userName := r.PostFormValue("username")
		password := r.PostFormValue("pass")

		log := kitlog.With(log.Get(r), "user", userName)

		level.Debug(log).
			Log("pass", password, "debug", "attempting auth")

		var token string

		err := a.q.RunTx(r.Context(), func(q *query.Queries) error {
			userData, err := q.GetAuthData(r.Context(), userName)
			if errors.Is(err, sql.ErrNoRows) {
				// Either no such user or no users at all
				count, err := q.CountUsers(r.Context())
				if err != nil {
					http.Error(w, err.Error(), http.StatusInternalServerError)
					return fmt.Errorf("getting user count: %w", err)
				}

				level.Debug(log).
					Log("num_users", count, "msg", "got user count")

				// TODO: Deal with this. Report an error if != 0 users in db, otherwise allow them through.

				w.WriteHeader(http.StatusUnauthorized)

				rd := responseData{
					Error: "authentication failed",
				}

				err = tpl.ExecuteTemplate(w, "auth.tpl", rd)
				if err != nil {
					return fmt.Errorf("executing auth template: %w", err)
				}

				return nil
			}
			if err != nil {
				return fmt.Errorf("getting salted password: %w", err)
			}

			hashed, err := hashPassword(r.Context(), q, userName, password)
			if err != nil {
				http.Error(w, err.Error(), http.StatusInternalServerError)
				return fmt.Errorf("hashing password: %w", err)
			}

			if hashed != userData.Password {
				level.Error(log).
					Log("msg", "password mismatch")

				w.WriteHeader(http.StatusUnauthorized)

				rd := responseData{
					Error: "authentication failed",
				}

				err := tpl.ExecuteTemplate(w, "auth.tpl", rd)
				if err != nil {
					return fmt.Errorf("executing auth template: %w", err)
				}

				return nil
			}

			log = kitlog.With(log, "user_id", userData.UserID)

			// All good
			level.Info(log).
				Log("msg", "auth looks good")

			addr := r.Header.Get("X-Forwarded-For")
			if addr == "" {
				addr = r.RemoteAddr
			}

			token, err = a.createSession(r.Context(), q, int(userData.UserID), addr)
			if err != nil {
				http.Error(w, err.Error(), http.StatusInternalServerError)
				return fmt.Errorf("creating session token: %w", err)
			}

			return nil
		})
		if err != nil {
			level.Error(log).Log("msg", "authentication failed", "error", err)

			// Response already written to client.
			return
		}

		cookie := http.Cookie{
			Name:     "session",
			Value:    token,
			Path:     "/",
			HttpOnly: true, // Might need to be unset if we do a REST-style API
		}

		level.Debug(log).
			Log("cookie", cookie, "msg", "setting token cookie")

		http.SetCookie(w, &cookie)

		// Redirect to initially requested URL
		http.Redirect(w, r, r.URL.String(), http.StatusFound)
	})
}

func (a Provider) createSession(ctx context.Context, q *query.Queries, userID int, remoteAddr string) (string, error) {
	var rawToken [40]byte

	n, err := crand.Read(rawToken[:])
	if err != nil {
		return "", err
	}

	if n != len(rawToken) {
		return "", errors.New("short read")
	}

	token := fmt.Sprintf("%02x", rawToken)

	err = q.InsertSession(ctx, query.InsertSessionParams{
		UserID:  int32(userID),
		Token:   token,
		Remote:  remoteAddr,
		Created: time.Now().Format(time.RFC3339Nano),
	})
	if err != nil {
		return "", err
	}

	return token, nil
}

func (a Provider) DeleteSession(ctx context.Context, token string) error {
	return a.q.DeleteSession(ctx, token)
}

func (a Provider) ListSessions(ctx context.Context) ([]Info, error) {
	rows, err := a.q.ListSessions(ctx)
	if err != nil {
		return nil, err
	}

	var sessions []Info
	for _, r := range rows {
		t, err := time.Parse(time.RFC3339Nano, r.Created)
		if err != nil {
			return nil, err
		}

		sessions = append(sessions, Info{
			Name:    r.Name,
			Created: t,
			Remote:  r.Remote,
		})
	}

	return sessions, nil
}

func (a Provider) ListUsers(ctx context.Context) ([]string, error) {
	return a.q.ListUsers(ctx)
}

func (a Provider) UpdatePassword(ctx context.Context, userName, passOld, passNew string, createNew bool) error {
	err := a.q.RunTx(ctx, func(q *query.Queries) error {
		hashedNew, err := hashPassword(ctx, q, userName, passNew)
		if err != nil {
			return err
		}

		if createNew {
			level.Info(log.GetContext(ctx)).
				Log("name", userName, "msg", "creating new user")

			err = q.InsertUser(ctx, query.InsertUserParams{
				Name:     userName,
				Password: hashedNew,
			})
			if err != nil {
				return err
			}

			return nil
		}

		authData, err := q.GetAuthData(ctx, userName)
		if err != nil {
			return err
		}

		hashedOld, err := hashPassword(ctx, q, userName, passOld)
		if err != nil {
			return err
		}

		if len(authData.Password) != 0 && authData.Password != hashedOld {
			// A password exists already and it's not the one the user gave as their 'old' password.
			return auth.ErrAuthFailed
		}

		level.Info(log.GetContext(ctx)).
			Log("name", userName, "msg", "updating password")

			// Either user doesn't exist yet (no entry) or the password they gave as their old matched. We
			// can update.
			// Try an update first, if that fails, an insert
		res, err := q.UpdateUser(ctx, query.UpdateUserParams{
			Name:     userName,
			Password: hashedNew,
		})

		n, err := res.RowsAffected()
		if err != nil {
			return err
		}
		if n > 1 {
			return fmt.Errorf("more than one user updated: %d", n)
		}
		if n > 0 {
			return nil
		}

		return nil
	})
	if err != nil {
		level.Error(log.GetContext(ctx)).
			Log("msg", "user update failed", "error", err)
		return err
	}

	return nil
}

func (a Provider) CreateUser(ctx context.Context, userName string) (string, error) {
	// generate random password
	const chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"

	var genPw string

	for idx := 0; idx < 16; idx++ {
		r := chars[rand.Intn(len(chars))]

		genPw += string(r)
	}

	err := a.UpdatePassword(ctx, userName, "", genPw, true)
	if err != nil {
		return "", fmt.Errorf("updating password for %q: %w", userName, err)
	}

	return genPw, nil
}

func (a Provider) DeleteUser(ctx context.Context, name string) error {
	return a.q.DeleteUser(ctx, name)
}
