package storage

import (
	"bytes"
	"context"
	"database/sql"
	"embed"
	"fmt"
	"image"
	"image/png"
	"io"
	"io/fs"
	"sort"

	"git.c3pb.de/gbe/invinoveritas/log"
	"git.c3pb.de/gbe/invinoveritas/storage/query"
	kitlog "github.com/go-kit/kit/log"
	"github.com/go-kit/kit/log/level"
	"golang.org/x/image/draw"
)

// AddPicture loads picture data (PNG or JPEG) from fh and persists a scaled down version in q.
// If something goes wrong during loading, or the image is neither PNG nor JPEG, an error
// is returned. If contentType is not the empty string, it is validated to be either
// image/png or image/jpeg.
func AddPicture(ctx context.Context, q *query.Queries, wineID int, fh io.Reader, contentType string) (err error) {
	// Encode scaled image as PNG, will contain image data if there is any
	switch contentType {
	case "", "image/jpeg", "image/png":
	default:
		return fmt.Errorf("unexpected content type for image: %q", contentType)
	}

	img, _, err := image.Decode(fh)
	if err != nil {
		return fmt.Errorf("decoding image: %w", err)
	}

	// Scale image down and encode as PNG
	// Get aspect ratio of incoming picture
	bounds := img.Bounds()
	aspect := float64(bounds.Max.X) / float64(bounds.Max.Y)

	const destHeight = 800

	rect := image.Rect(0, 0, int(destHeight*aspect), destHeight)

	level.Info(log.GetContext(ctx)).
		Log("bounds", bounds,
			"aspect", aspect,
			"rect", rect,
			"msg", "resizing image")

	scaled := image.NewRGBA(rect)

	draw.ApproxBiLinear.Scale(scaled, rect, img, bounds, draw.Over, nil)

	var buf bytes.Buffer
	err = png.Encode(&buf, scaled)
	if err != nil {
		return fmt.Errorf("encoding image: %w", err)
	}

	err = q.StorePicture(ctx, query.StorePictureParams{
		WineID:  int32(wineID),
		Picture: buf.Bytes(),
	})
	if err != nil {
		return fmt.Errorf("storing image: %w", err)
	}

	return nil
}

//go:embed migrations/*.sql
var migrationFS embed.FS

func Open(ctx context.Context, path string, logger kitlog.Logger) (*sql.DB, error) {
	db, err := sql.Open("sqlite", path)
	if err != nil {
		return nil, err
	}

	level.Debug(logger).Log("db_path", path)

	pragmas := []string{
		"PRAGMA busy_timeout = 600000",
		"PRAGMA journal_mode = WAL",
	}

	for _, p := range pragmas {
		_, err = db.ExecContext(ctx, p)
		if err != nil {
			return nil, fmt.Errorf("running %q: %w", p, err)
		}
	}

	// Make sure that we have a table for the migrations, the rest runs in a transaction.
	_, err = db.ExecContext(ctx, `CREATE TABLE IF NOT EXISTS migrations (name text, unique(name))`)
	if err != nil {
		return nil, fmt.Errorf("creating migrations table: %w", err)
	}

	q := query.New(db)

	entries, err := migrationFS.ReadDir("migrations")
	if err != nil {
		return nil, fmt.Errorf("reading embedded migration FS: %w", err)
	}

	sort.Slice(entries, func(i, j int) bool {
		return entries[i].Name() < entries[j].Name()
	})

	err = q.RunTx(ctx, func(q *query.Queries) error {
		for _, e := range entries {
			// Check if we need to apply that migration
			count, err := q.CountMigrations(ctx, e.Name())
			if err != nil {
				return fmt.Errorf("checking for migration %q: %w", e.Name(), err)
			}

			if count > 1 {
				return fmt.Errorf("migration %q applied more than once", e.Name())
			}

			if count == 1 {
				continue
			}

			content, err := fs.ReadFile(migrationFS, "migrations/"+e.Name())
			if err != nil {
				return fmt.Errorf("reading migration %q: %w", e.Name(), err)
			}

			_, err = q.Raw().ExecContext(ctx, string(content))
			if err != nil {
				return fmt.Errorf("applying migration %q: %w", e.Name(), err)
			}

			// Mark migration as applied
			err = q.AddMigration(ctx, e.Name())
			if err != nil {
				return fmt.Errorf("marking migration %q as applied: %w", e.Name(), err)
			}
		}

		return nil
	})
	if err != nil {
		return nil, fmt.Errorf("applying migrations: %w", err)
	}

	return db, nil
}
